from setuptools import setup, find_packages

setup(name='recpulse',
      version='0.0.11',
      description='ML framework',
      author='Maksym Petrenko',
      author_email='maksym.petrenko.a@gmail.com',
      packages=['recpulse'],
      install_requires=[

      ],
)
